import json
from typing import Any

from grpc import Channel
from tblib import pickling_support

pickling_support.install()
import pickle

from .serpent_pb2 import ArgsKwargs, MethodCall, Response, Result, Error, SetupResponse, Empty, Attr, AttrValue
from .serpent_pb2_grpc import SerpentServicer, SerpentStub


class SerpentServicer(SerpentServicer):

    def __init__(self, inner: Any):
        self.inner = inner

    @property
    def gettables(self):
        return [attr for attr in dir(self.inner) if not callable(getattr(self, attr))]

    def Setup(self, request, context):
        return SetupResponse(
            attributes=[attr for attr in dir(self.inner) if not callable(getattr(self.inner, attr)) and not attr.startswith("__")],
            methods=[meth for meth in dir(self.inner) if callable(getattr(self.inner, meth)) and not meth.startswith("__")]
        )

    def Call(self, request: MethodCall, context: Any) -> Response:
        try:
            args = json.loads(request.argsKwargs.args)
            kwargs = json.loads(request.argsKwargs.kwargs)
            result = getattr(self.inner, request.method)(*args, **kwargs)
            return Response(result=Result(value=json.dumps(result)))
        except Exception as e:
            return Response(error=Error(data=pickle.dumps(e)))

    def GetAttr(self, request: Attr, context):
        value = getattr(self.inner, request.name)
        value = json.dumps(value)
        return Response(result=Result(value=value))

    def SetAttr(self, request: AttrValue, context):
        setattr(self.inner, request.name, json.loads(request.value))
        return Response()


class SerpentClient(SerpentStub):
    def __init__(self, channel: Channel):
        super().__init__(channel)

        setup: SetupResponse = self.Setup(Empty())
        self.__attributes = setup.attributes
        self.__methods = setup.methods

        for attr in self.__attributes:
            self.__add_property(attr)

        for meth in self.__methods:
            self.__add_method(meth)

    def __add_property(self, name):
        def getter(self):
            resp = self.GetAttr(Attr(name=name))
            value = json.loads(resp.result.value)
            return value

        def setter(self, value):
            self.SetAttr(AttrValue(name=name, value=json.dumps(value)))

        setattr(self.__class__, name, property(getter, setter))

    def __add_method(self, name):
        def method(*args: Any, **kwargs: Any) -> Any:
            args = json.dumps(args)
            kwargs = json.dumps(kwargs)
            method_call_data = MethodCall(
                method=name,
                argsKwargs=ArgsKwargs(args=args, kwargs=kwargs))

            response: Response = self.Call(method_call_data)

            if response.HasField("result"):
                return json.loads(response.result.value)
            raise pickle.loads(response.error.data)
        setattr(self, name, method)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: serpent_serve/serpent.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'serpent_serve/serpent.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bserpent_serve/serpent.proto\x12\x07serpent\"\x07\n\x05\x45mpty\"H\n\nArgsKwargs\x12\x11\n\x04\x61rgs\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x13\n\x06kwargs\x18\x02 \x01(\tH\x01\x88\x01\x01\x42\x07\n\x05_argsB\t\n\x07_kwargs\"E\n\nMethodCall\x12\x0e\n\x06method\x18\x01 \x01(\t\x12\'\n\nargsKwargs\x18\x02 \x01(\x0b\x32\x13.serpent.ArgsKwargs\"&\n\x06Result\x12\x12\n\x05value\x18\x01 \x01(\tH\x00\x88\x01\x01\x42\x08\n\x06_value\"\x15\n\x05\x45rror\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\"Z\n\x08Response\x12!\n\x06result\x18\x01 \x01(\x0b\x32\x0f.serpent.ResultH\x00\x12\x1f\n\x05\x65rror\x18\x02 \x01(\x0b\x32\x0e.serpent.ErrorH\x00\x42\n\n\x08response\"4\n\rSetupResponse\x12\x12\n\nattributes\x18\x01 \x03(\t\x12\x0f\n\x07methods\x18\x02 \x03(\t\"\x14\n\x04\x41ttr\x12\x0c\n\x04name\x18\x01 \x01(\t\"(\n\tAttrValue\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t2\xd1\x01\n\x07Serpent\x12\x31\n\x05Setup\x12\x0e.serpent.Empty\x1a\x16.serpent.SetupResponse\"\x00\x12\x30\n\x04\x43\x61ll\x12\x13.serpent.MethodCall\x1a\x11.serpent.Response\"\x00\x12-\n\x07GetAttr\x12\r.serpent.Attr\x1a\x11.serpent.Response\"\x00\x12\x32\n\x07SetAttr\x12\x12.serpent.AttrValue\x1a\x11.serpent.Response\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'serpent_serve.serpent_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_EMPTY']._serialized_start=40
  _globals['_EMPTY']._serialized_end=47
  _globals['_ARGSKWARGS']._serialized_start=49
  _globals['_ARGSKWARGS']._serialized_end=121
  _globals['_METHODCALL']._serialized_start=123
  _globals['_METHODCALL']._serialized_end=192
  _globals['_RESULT']._serialized_start=194
  _globals['_RESULT']._serialized_end=232
  _globals['_ERROR']._serialized_start=234
  _globals['_ERROR']._serialized_end=255
  _globals['_RESPONSE']._serialized_start=257
  _globals['_RESPONSE']._serialized_end=347
  _globals['_SETUPRESPONSE']._serialized_start=349
  _globals['_SETUPRESPONSE']._serialized_end=401
  _globals['_ATTR']._serialized_start=403
  _globals['_ATTR']._serialized_end=423
  _globals['_ATTRVALUE']._serialized_start=425
  _globals['_ATTRVALUE']._serialized_end=465
  _globals['_SERPENT']._serialized_start=468
  _globals['_SERPENT']._serialized_end=677
# @@protoc_insertion_point(module_scope)



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Empty(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ArgsKwargs(_message.Message):
    __slots__ = ("args", "kwargs")
    ARGS_FIELD_NUMBER: _ClassVar[int]
    KWARGS_FIELD_NUMBER: _ClassVar[int]
    args: str
    kwargs: str
    def __init__(self, args: _Optional[str] = ..., kwargs: _Optional[str] = ...) -> None: ...

class MethodCall(_message.Message):
    __slots__ = ("method", "argsKwargs")
    METHOD_FIELD_NUMBER: _ClassVar[int]
    ARGSKWARGS_FIELD_NUMBER: _ClassVar[int]
    method: str
    argsKwargs: ArgsKwargs
    def __init__(self, method: _Optional[str] = ..., argsKwargs: _Optional[_Union[ArgsKwargs, _Mapping]] = ...) -> None: ...

class Result(_message.Message):
    __slots__ = ("value",)
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: str
    def __init__(self, value: _Optional[str] = ...) -> None: ...

class Error(_message.Message):
    __slots__ = ("data",)
    DATA_FIELD_NUMBER: _ClassVar[int]
    data: bytes
    def __init__(self, data: _Optional[bytes] = ...) -> None: ...

class Response(_message.Message):
    __slots__ = ("result", "error")
    RESULT_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    result: Result
    error: Error
    def __init__(self, result: _Optional[_Union[Result, _Mapping]] = ..., error: _Optional[_Union[Error, _Mapping]] = ...) -> None: ...

class SetupResponse(_message.Message):
    __slots__ = ("attributes", "methods")
    ATTRIBUTES_FIELD_NUMBER: _ClassVar[int]
    METHODS_FIELD_NUMBER: _ClassVar[int]
    attributes: _containers.RepeatedScalarFieldContainer[str]
    methods: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, attributes: _Optional[_Iterable[str]] = ..., methods: _Optional[_Iterable[str]] = ...) -> None: ...

class Attr(_message.Message):
    __slots__ = ("name",)
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class AttrValue(_message.Message):
    __slots__ = ("name", "value")
    NAME_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    name: str
    value: str
    def __init__(self, name: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from serpent_serve import serpent_pb2 as serpent__serve_dot_serpent__pb2

GRPC_GENERATED_VERSION = '1.68.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in serpent_serve/serpent_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class SerpentStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Setup = channel.unary_unary(
                '/serpent.Serpent/Setup',
                request_serializer=serpent__serve_dot_serpent__pb2.Empty.SerializeToString,
                response_deserializer=serpent__serve_dot_serpent__pb2.SetupResponse.FromString,
                _registered_method=True)
        self.Call = channel.unary_unary(
                '/serpent.Serpent/Call',
                request_serializer=serpent__serve_dot_serpent__pb2.MethodCall.SerializeToString,
                response_deserializer=serpent__serve_dot_serpent__pb2.Response.FromString,
                _registered_method=True)
        self.GetAttr = channel.unary_unary(
                '/serpent.Serpent/GetAttr',
                request_serializer=serpent__serve_dot_serpent__pb2.Attr.SerializeToString,
                response_deserializer=serpent__serve_dot_serpent__pb2.Response.FromString,
                _registered_method=True)
        self.SetAttr = channel.unary_unary(
                '/serpent.Serpent/SetAttr',
                request_serializer=serpent__serve_dot_serpent__pb2.AttrValue.SerializeToString,
                response_deserializer=serpent__serve_dot_serpent__pb2.Response.FromString,
                _registered_method=True)


class SerpentServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Setup(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Call(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetAttr(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetAttr(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_SerpentServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Setup': grpc.unary_unary_rpc_method_handler(
                    servicer.Setup,
                    request_deserializer=serpent__serve_dot_serpent__pb2.Empty.FromString,
                    response_serializer=serpent__serve_dot_serpent__pb2.SetupResponse.SerializeToString,
            ),
            'Call': grpc.unary_unary_rpc_method_handler(
                    servicer.Call,
                    request_deserializer=serpent__serve_dot_serpent__pb2.MethodCall.FromString,
                    response_serializer=serpent__serve_dot_serpent__pb2.Response.SerializeToString,
            ),
            'GetAttr': grpc.unary_unary_rpc_method_handler(
                    servicer.GetAttr,
                    request_deserializer=serpent__serve_dot_serpent__pb2.Attr.FromString,
                    response_serializer=serpent__serve_dot_serpent__pb2.Response.SerializeToString,
            ),
            'SetAttr': grpc.unary_unary_rpc_method_handler(
                    servicer.SetAttr,
                    request_deserializer=serpent__serve_dot_serpent__pb2.AttrValue.FromString,
                    response_serializer=serpent__serve_dot_serpent__pb2.Response.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'serpent.Serpent', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('serpent.Serpent', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Serpent(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Setup(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/serpent.Serpent/Setup',
            serpent__serve_dot_serpent__pb2.Empty.SerializeToString,
            serpent__serve_dot_serpent__pb2.SetupResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Call(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/serpent.Serpent/Call',
            serpent__serve_dot_serpent__pb2.MethodCall.SerializeToString,
            serpent__serve_dot_serpent__pb2.Response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetAttr(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/serpent.Serpent/GetAttr',
            serpent__serve_dot_serpent__pb2.Attr.SerializeToString,
            serpent__serve_dot_serpent__pb2.Response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SetAttr(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/serpent.Serpent/SetAttr',
            serpent__serve_dot_serpent__pb2.AttrValue.SerializeToString,
            serpent__serve_dot_serpent__pb2.Response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)
